/*
 * Test Class for Heart
 */
package ch.zhaw.prog2.heartbeat;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import ch.zhaw.prog2.heartbeat.parts.Valve;
import org.junit.Before;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;

import ch.zhaw.prog2.heartbeat.Heart.State;
import org.junit.rules.ExpectedException;
import org.junit.runner.RunWith;
import org.mockito.InOrder;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.mockito.junit.MockitoJUnitRunner;

@RunWith(MockitoJUnitRunner.class)
class HeartTest {
    /**
     * This is a very simple test to check if Junit and Mockito are properly set up.
     */
    @Test
    void testTheTest() {
        Heart classUnderTest = new Heart();
        assertNotNull(classUnderTest.getState(), "The heart must have a state.");
    }

    /**
     * Tests a single heartbeat
     */
    @Test
    void testHeartBeat() throws Heart.HeartBeatDysfunctionException, Valve.IllegalValveStateException, Heart.InvalidValvePositionException {
        Heart heart = new Heart();
        State startState = heart.getState();

        heart.executeHeartBeat();

        // after one heartbeat, the heart must be in the same state as before
        assertEquals(startState, heart.getState());
    }

    /**
     * Tests if the valves are open or closed depending on the status of the heart
     */
    @Test
    void testValveStatus() throws Heart.HeartBeatDysfunctionException, Valve.IllegalValveStateException, Heart.InvalidValvePositionException {
        Heart heart = new Heart();

        heart.executeHeartBeat();

        State state = heart.getState();

        if (state.equals(Heart.State.DIASTOLE)) {
            for (Half half : heart.getHalves()) {
                assertFalse(half.isAtrioventricularValveOpen());
                assertTrue(half.isSemilunarValveOpen());
            }
        } else if ((state.equals(Heart.State.SYSTOLE))) {
            for (Half half : heart.getHalves()) {
                assertTrue(half.isAtrioventricularValveOpen());
                assertFalse(half.isSemilunarValveOpen());
            }
        }
    }

    /**
     * Tests if the hart throws the appropriate Exception, when malfunction was detected during hartBeat
     */
    @Test
    void testExecuteHeartBeatErrorBehaviour() {
        Heart heart = new Heart();
        // prepare error situation due to wrong initialization
        heart.setState(State.SYSTOLE);

        assertThrows(Heart.HeartBeatDysfunctionException .class, // verification using lambda
            () -> heart.executeHeartBeat());
    }

    /**
     * Tests if the hart throws the appropriate Exception, when malfunction was detected during hartBeat
     * with exception Stubbing
     */
    @Test
    void testExecuteHartBeatErrorBehaviourWithStubbing() {
        //TODO implement
        Half half = spy(new Half(Half.Side.LEFT));
        Heart heart = new Heart(half, new Half(Half.Side.RIGHT));
        half.initializeState(heart.getState());
        when(half.isAtrioventricularValveOpen()).thenReturn(false);

        assertThrows(Heart.HeartBeatDysfunctionException .class, // verification using lambda
            () -> heart.executeHeartBeat());
    }

    /**
     * We test if Heart::executeHeartbeat() sends the right signals to both of its
     * halves.
     *
     * When Half::contractVentricle() is called, Half::closeAtrioventricularValve()
     * and Half::openSemilunarValve() must have been called earlier.
     *
     */
    @Test
    void testValvesBehavior() throws Valve.IllegalValveStateException, Heart.HeartBeatDysfunctionException {
        //TODO implement
        Half halfLeft = spy(new Half(Half.Side.LEFT));
        Half halfRight = spy(new Half(Half.Side.RIGHT));
        Heart heart = new Heart(halfLeft, halfRight);

        heart.executeHeartBeat();

        InOrder inOrderHalfLeft = inOrder(halfLeft);

        inOrderHalfLeft.verify(halfLeft).openAtrioventricularValve();
        inOrderHalfLeft.verify(halfLeft).closeSemilunarValve();
        inOrderHalfLeft.verify(halfLeft).relaxAtrium();
        inOrderHalfLeft.verify(halfLeft).relaxVentricle();

        inOrderHalfLeft.verify(halfLeft).closeAtrioventricularValve();
        inOrderHalfLeft.verify(halfLeft).openSemilunarValve();
        inOrderHalfLeft.verify(halfLeft).contractAtrium();
        inOrderHalfLeft.verify(halfLeft).contractVentricle();

        inOrderHalfLeft.verifyNoMoreInteractions();

        InOrder inOrderHalfRight = inOrder(halfRight);

        inOrderHalfRight.verify(halfRight).openAtrioventricularValve();
        inOrderHalfRight.verify(halfRight).closeSemilunarValve();
        inOrderHalfRight.verify(halfRight).relaxAtrium();
        inOrderHalfRight.verify(halfRight).relaxVentricle();

        inOrderHalfRight.verify(halfRight).closeAtrioventricularValve();
        inOrderHalfRight.verify(halfRight).openSemilunarValve();
        inOrderHalfRight.verify(halfRight).contractAtrium();
        inOrderHalfRight.verify(halfRight).contractVentricle();

        inOrderHalfRight.verifyNoMoreInteractions();

    }

    @Test
    void testDiastoleExceptionAtrioventricularValve(){
        Half half = spy(new Half(Half.Side.LEFT));
        Heart heart = new Heart(half, new Half(Half.Side.RIGHT));
        when(half.isAtrioventricularValveOpen()).thenReturn(true);
        assertThrows(Heart.InvalidValvePositionException .class, // verification using lambda
            () -> heart.executeDiastole());
    }

    @Test
    void testDiastoleExceptionSemilunarValve(){
        Half half = spy(new Half(Half.Side.LEFT));
        Heart heart = new Heart(half, new Half(Half.Side.RIGHT));
        when(half.isSemilunarValveOpen()).thenReturn(false);
        assertThrows(Heart.InvalidValvePositionException .class, // verification using lambda
            () -> heart.executeDiastole());
    }

    @Test
    void testSystoleExceptionAtrioventricularValve(){
        Half half = spy(new Half(Half.Side.LEFT));
        Heart heart = new Heart(half, new Half(Half.Side.RIGHT));
        when(half.isAtrioventricularValveOpen()).thenReturn(false);
        assertThrows(Heart.InvalidValvePositionException .class, // verification using lambda
            () -> heart.executeSystole());
    }

    @Test
    void testSystoleExceptionSemilunarValve(){
        Half half = spy(new Half(Half.Side.LEFT));
        Heart heart = new Heart(half, new Half(Half.Side.RIGHT));
        when(half.isSemilunarValveOpen()).thenReturn(true);
        assertThrows(Heart.InvalidValvePositionException .class, // verification using lambda
            () -> heart.executeSystole());
    }
}
